use std::collections::BTreeMap;
use std::fmt::{Error, Write};
use std::iter::zip;

use hercules_ir::*;

use crate::*;

/*
 * Entry Hercules functions are lowered to async Rust code to achieve easy task
 * level parallelism. This Rust is generated textually, and is included via a
 * procedural macro in the user's Rust code.
 *
 * Generating Rust that properly handles memory across devices is tricky. In
 * particular, the following elements are challenges:
 *
 * 1. RT functions may return objects that were not first passed in as
 *    parameters, via object constants.
 * 2. We want to allocate as much memory upfront as possible. Our goal is for
 *    each call to a RT function from host Rust code, there is at most one
 *    memory allocation per device.
 * 3. We want to statically determine when cross-device communication is
 *    necessary - this should be a separate concern from allocation.
 * 4. At the boundary between host Rust / RT functions, we want to encode
 *    lifetime rules so that the Hercules API is memory safe.
 * 5. We want to support efficient composition of Hercules RT functions in both
 *    synchronous and asynchronous contexts.
 *
 * Challenges #1 and #2 require that RT functions themselves do not allocate
 * memory. Instead, for every entry point function, a "runner" type will be
 * generated. The host Rust code must instantiate a runner object to invoke an
 * entry point function. This runner object contains a "backing" memory that is
 * the single allocation of memories for this function. The runner object can be
 * used to call the same entry point multiple times, and to the extent possible
 * the backing memory will be re-used. The size of the backing memory depends on
 * the dynamic constants passed in to the entry point, so it's lazily allocated
 * on calls to the entry point to the needed size. To address challenge #4, any
 * returned objects will be lifetime-bound to the runner object instance,
 * ensuring that the reference cannot be used after the runner object has de-
 * allocated the backing memory. This also ensures that the runner can't be run
 * again while a returned object from a previous iteration is still live, since
 * the entry point method requires an exclusive reference to the runner.
 *
 * Addressing challenge #3 requires we determine what objects are live on what
 * devices at what times. This can be done fairly easily by coloring nodes by
 * what device they produce their result on and inserting inter-device transfers
 * along edges connecting nodes of different colors. Nodes can only have a
 * single color - this is enforced by the GCM pass.
 *
 * Addressing challenge #5 requires runner objects for entry points accept and
 * return objects that are not in their own backing memory and potentially on
 * any device. For this reason, parameter and return nodes are not necessarily
 * CPU colored. Instead, runners take and return Hercules reference objects that
 * refer to memory on some device which have unknown origin. Hercules reference
 * objects have a lifetime parameter, and when a runner may return a Hercules
 * reference that refers to its backing memory, the lifetime of the Hercules
 * reference is the same as the lifetime of the mutable reference of the runner
 * used in the entry point signature. In other words, the RT backend infers the
 * proper lifetime bounds on parameter and returned Hercules reference objects
 * in relation to the runner's self reference using the collection objects
 * analysis. There are the following kinds of Hercules reference objects:
 *
 * - HerculesCPURef
 * - HerculesCPURefMut
 * - HerculesCUDARef
 * - HerculesCUDARefMut
 *
 * Essentially, there are types for each device, one for immutable references
 * and one for exclusive references. Mutable references can decay into immutable
 * references, and immutable references can be cloned. The CPU reference types
 * can be created from normal Rust references. The CUDA reference types can't be
 * created from normal Rust references - for that purpose, an additional type is
 * given, CUDABox, which essentially allows the user to manually allocate and
 * set some CUDA memory - the user can then take a CUDA reference to that box.
 */
pub fn rt_codegen<W: Write>(
    func_id: FunctionID,
    module: &Module,
    typing: &Vec<TypeID>,
    control_subgraph: &Subgraph,
    collection_objects: &CollectionObjects,
    callgraph: &CallGraph,
    devices: &Vec<Device>,
    bbs: &BasicBlocks,
    node_colors: &FunctionNodeColors,
    backing_allocations: &BackingAllocations,
    w: &mut W,
) -> Result<(), Error> {
    let ctx = RTContext {
        func_id,
        module,
        typing,
        control_subgraph,
        collection_objects,
        callgraph,
        devices,
        bbs,
        node_colors,
        backing_allocations,
    };
    ctx.codegen_function(w)
}

struct RTContext<'a> {
    func_id: FunctionID,
    module: &'a Module,
    typing: &'a Vec<TypeID>,
    control_subgraph: &'a Subgraph,
    collection_objects: &'a CollectionObjects,
    callgraph: &'a CallGraph,
    devices: &'a Vec<Device>,
    bbs: &'a BasicBlocks,
    node_colors: &'a FunctionNodeColors,
    backing_allocations: &'a BackingAllocations,
}

impl<'a> RTContext<'a> {
    fn codegen_function<W: Write>(&self, w: &mut W) -> Result<(), Error> {
        // If this is an entry function, generate a corresponding runner object
        // type definition.
        let func = &self.get_func();
        if func.entry {
            self.codegen_runner_object(w)?;
        }

        // Dump the function signature.
        write!(
            w,
            "#[allow(unused_assignments,unused_variables,unused_mut,unused_parens,unused_unsafe,non_snake_case)]\nasync unsafe fn {}(",
            func.name
        )?;
        let mut first_param = true;
        // The first set of parameters are pointers to backing memories.
        for (device, _) in self.backing_allocations[&self.func_id].iter() {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "backing_{}: *mut u8", device.name())?;
        }
        // The second set of parameters are dynamic constants.
        for idx in 0..func.num_dynamic_constants {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "dc_p{}: u64", idx)?;
        }
        // The third set of parameters are normal parameters.
        for idx in 0..func.param_types.len() {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "p{}: {}", idx, self.get_type(func.param_types[idx]))?;
        }
        write!(w, ") -> {} {{\n", self.get_type(func.return_type))?;

        // Dump signatures for called device functions.
        write!(w, "    extern \"C\" {{\n")?;
        for callee_id in self.callgraph.get_callees(self.func_id) {
            if self.devices[callee_id.idx()] == Device::AsyncRust {
                continue;
            }
            let callee = &self.module.functions[callee_id.idx()];
            write!(w, "        fn {}(", callee.name)?;
            let mut first_param = true;
            if self.backing_allocations[&callee_id].contains_key(&self.devices[callee_id.idx()]) {
                first_param = false;
                write!(w, "backing: *mut u8")?;
            }
            for idx in 0..callee.num_dynamic_constants {
                if first_param {
                    first_param = false;
                } else {
                    write!(w, ", ")?;
                }
                write!(w, "dc{}: u64", idx)?;
            }
            for (idx, ty) in callee.param_types.iter().enumerate() {
                if first_param {
                    first_param = false;
                } else {
                    write!(w, ", ")?;
                }
                write!(w, "p{}: {}", idx, self.get_type(*ty))?;
            }
            write!(w, ") -> {};\n", self.get_type(callee.return_type))?;
        }
        write!(w, "    }}\n")?;

        // Declare intermediary variables for every value.
        for idx in 0..func.nodes.len() {
            if func.nodes[idx].is_control() {
                continue;
            }
            write!(
                w,
                "    let mut node_{}: {} = {};\n",
                idx,
                self.get_type(self.typing[idx]),
                if self.module.types[self.typing[idx].idx()].is_integer() {
                    "0"
                } else if self.module.types[self.typing[idx].idx()].is_float() {
                    "0.0"
                } else {
                    "::core::ptr::null_mut()"
                }
            )?;
        }

        // The core executor is a Rust loop. We literally run a "control token"
        // as described in the original sea of nodes paper through the basic
        // blocks to drive execution.
        write!(
            w,
            "    let mut control_token: i8 = 0;\n    loop {{\n        match control_token {{\n",
        )?;

        let mut blocks: BTreeMap<_, _> = (0..func.nodes.len())
            .filter(|idx| func.nodes[*idx].is_control())
            .map(|idx| (NodeID::new(idx), String::new()))
            .collect();

        // Emit data flow into basic blocks.
        for block in self.bbs.1.iter() {
            for id in block {
                self.codegen_data_node(*id, &mut blocks)?;
            }
        }

        // Emit control flow into basic blocks.
        for id in (0..func.nodes.len()).map(NodeID::new) {
            if !func.nodes[id.idx()].is_control() {
                continue;
            }
            self.codegen_control_node(id, &mut blocks)?;
        }

        // Dump the emitted basic blocks.
        for (id, block) in blocks {
            write!(
                w,
                "            {} => {{\n{}            }}\n",
                id.idx(),
                block
            )?;
        }

        // Close the match and loop.
        write!(w, "            _ => panic!()\n        }}\n    }}\n")?;
        write!(w, "}}\n")?;
        Ok(())
    }

    /*
     * While control nodes in Hercules IR are predecessor-centric (each take a
     * control input that defines the predecessor relationship), the Rust loop
     * we generate is successor centric. This difference requires explicit
     * translation.
     */
    fn codegen_control_node(
        &self,
        id: NodeID,
        blocks: &mut BTreeMap<NodeID, String>,
    ) -> Result<(), Error> {
        let func = &self.get_func();
        match func.nodes[id.idx()] {
            // Start, region, and projection control nodes all have exactly one
            // successor and are otherwise simple.
            Node::Start
            | Node::Region { preds: _ }
            | Node::Projection {
                control: _,
                selection: _,
            } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                let succ = self.control_subgraph.succs(id).next().unwrap();
                write!(block, "                control_token = {};\n", succ.idx())?
            }
            // If nodes have two successors - examine the projections to
            // determine which branch is which, and branch between them.
            Node::If { control: _, cond } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                let mut succs = self.control_subgraph.succs(id);
                let succ1 = succs.next().unwrap();
                let succ2 = succs.next().unwrap();
                let succ1_is_true = func.nodes[succ1.idx()].try_projection(1).is_some();
                write!(
                    block,
                    "                control_token = if {} {{ {} }} else {{ {} }};\n",
                    self.get_value(cond),
                    if succ1_is_true { succ1 } else { succ2 }.idx(),
                    if succ1_is_true { succ2 } else { succ1 }.idx(),
                )?
            }
            Node::Return { control: _, data } => {
                let block = &mut blocks.get_mut(&id).unwrap();
                write!(block, "                return {};\n", self.get_value(data))?
            }
            _ => panic!("PANIC: Can't lower {:?}.", func.nodes[id.idx()]),
        }
        Ok(())
    }

    /*
     * Lower data nodes in Hercules IR into Rust statements.
     */
    fn codegen_data_node(
        &self,
        id: NodeID,
        blocks: &mut BTreeMap<NodeID, String>,
    ) -> Result<(), Error> {
        let func = &self.get_func();
        match func.nodes[id.idx()] {
            Node::Parameter { index } => {
                let block = &mut blocks.get_mut(&self.bbs.0[id.idx()]).unwrap();
                write!(
                    block,
                    "                {} = p{};\n",
                    self.get_value(id),
                    index
                )?
            }
            Node::Constant { id: cons_id } => {
                let block = &mut blocks.get_mut(&self.bbs.0[id.idx()]).unwrap();
                write!(block, "                {} = ", self.get_value(id))?;
                let mut size_and_device = None;
                match self.module.constants[cons_id.idx()] {
                    Constant::Boolean(val) => write!(block, "{}bool", val)?,
                    Constant::Integer8(val) => write!(block, "{}i8", val)?,
                    Constant::Integer16(val) => write!(block, "{}i16", val)?,
                    Constant::Integer32(val) => write!(block, "{}i32", val)?,
                    Constant::Integer64(val) => write!(block, "{}i64", val)?,
                    Constant::UnsignedInteger8(val) => write!(block, "{}u8", val)?,
                    Constant::UnsignedInteger16(val) => write!(block, "{}u16", val)?,
                    Constant::UnsignedInteger32(val) => write!(block, "{}u32", val)?,
                    Constant::UnsignedInteger64(val) => write!(block, "{}u64", val)?,
                    Constant::Float32(val) => write!(block, "{}f32", val)?,
                    Constant::Float64(val) => write!(block, "{}f64", val)?,
                    Constant::Product(ty, _)
                    | Constant::Summation(ty, _, _)
                    | Constant::Array(ty) => {
                        let (device, offset) = self.backing_allocations[&self.func_id]
                            .iter()
                            .filter_map(|(device, (_, offsets))| {
                                offsets.get(&id).map(|id| (*device, *id))
                            })
                            .next()
                            .unwrap();
                        write!(block, "backing_{}.byte_add(", device.name())?;
                        self.codegen_dynamic_constant(offset, block)?;
                        write!(block, " as usize)")?;
                        size_and_device = Some((self.codegen_type_size(ty), device));
                    }
                }
                write!(block, ";\n")?;
                if !func.schedules[id.idx()].contains(&Schedule::NoResetConstant) {
                    if let Some((size, device)) = size_and_device {
                        write!(
                            block,
                            "                ::hercules_rt::__{}_zero_mem({}, {} as usize);\n",
                            device.name(),
                            self.get_value(id),
                            size
                        )?;
                    }
                }
            }
            Node::Call {
                control: _,
                function: callee_id,
                ref dynamic_constants,
                ref args,
            } => {
                // The device backends ensure that device functions have the
                // same interface as AsyncRust functions.
                let block = &mut blocks.get_mut(&self.bbs.0[id.idx()]).unwrap();
                write!(
                    block,
                    "                {} = {}(",
                    self.get_value(id),
                    self.module.functions[callee_id.idx()].name
                )?;
                for (device, offset) in self.backing_allocations[&self.func_id]
                    .iter()
                    .filter_map(|(device, (_, offsets))| offsets.get(&id).map(|id| (*device, *id)))
                {
                    write!(block, "backing_{}.byte_add(", device.name())?;
                    self.codegen_dynamic_constant(offset, block)?;
                    write!(block, "), ")?
                }
                for dc in dynamic_constants {
                    self.codegen_dynamic_constant(*dc, block)?;
                    write!(block, ", ")?;
                }
                for arg in args {
                    write!(block, "{}, ", self.get_value(*arg))?;
                }
                let device = self.devices[callee_id.idx()];
                if device == Device::AsyncRust {
                    write!(block, ").await;\n")?;
                } else {
                    write!(block, ");\n")?;
                }
            }
            Node::Read {
                collect,
                ref indices,
            } => {
                let block = &mut blocks.get_mut(&self.bbs.0[id.idx()]).unwrap();
                let collect_ty = self.typing[collect.idx()];
                let out_size = self.codegen_type_size(self.typing[id.idx()]);
                let offset = self.codegen_index_math(collect_ty, indices)?;
                todo!();
            }
            Node::Write {
                collect,
                data,
                ref indices,
            } => {
                let block = &mut blocks.get_mut(&self.bbs.0[id.idx()]).unwrap();
                let collect_ty = self.typing[collect.idx()];
                let data_size = self.codegen_type_size(self.typing[data.idx()]);
                let offset = self.codegen_index_math(collect_ty, indices)?;
                todo!();
            }
            _ => panic!(
                "PANIC: Can't lower {:?} in {}.",
                func.nodes[id.idx()],
                func.name
            ),
        }
        Ok(())
    }

    /*
     * Lower dynamic constant in Hercules IR into a Rust expression.
     */
    fn codegen_dynamic_constant<W: Write>(
        &self,
        id: DynamicConstantID,
        w: &mut W,
    ) -> Result<(), Error> {
        match &self.module.dynamic_constants[id.idx()] {
            DynamicConstant::Constant(val) => write!(w, "{}", val)?,
            DynamicConstant::Parameter(idx) => write!(w, "dc_p{}", idx)?,
            DynamicConstant::Add(xs) => {
                write!(w, "(")?;
                let mut xs = xs.iter();
                self.codegen_dynamic_constant(*xs.next().unwrap(), w)?;
                for x in xs {
                    write!(w, "+")?;
                    self.codegen_dynamic_constant(*x, w)?;
                }
                write!(w, ")")?;
            }
            DynamicConstant::Sub(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "-")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Mul(xs) => {
                write!(w, "(")?;
                let mut xs = xs.iter();
                self.codegen_dynamic_constant(*xs.next().unwrap(), w)?;
                for x in xs {
                    write!(w, "*")?;
                    self.codegen_dynamic_constant(*x, w)?;
                }
                write!(w, ")")?;
            }
            DynamicConstant::Div(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "/")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Rem(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "%")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Min(xs) => {
                let mut xs = xs.iter().peekable();

                // Track the number of parentheses we open that need to be closed later
                let mut opens = 0;
                while let Some(x) = xs.next() {
                    if xs.peek().is_none() {
                        // For the last element, we just print it
                        self.codegen_dynamic_constant(*x, w)?;
                    } else {
                        // Otherwise, we create a new call to min and print the element as the
                        // first argument
                        write!(w, "::core::cmp::min(")?;
                        self.codegen_dynamic_constant(*x, w)?;
                        write!(w, ",")?;
                        opens += 1;
                    }
                }
                for _ in 0..opens {
                    write!(w, ")")?;
                }
            }
            DynamicConstant::Max(xs) => {
                let mut xs = xs.iter().peekable();

                let mut opens = 0;
                while let Some(x) = xs.next() {
                    if xs.peek().is_none() {
                        self.codegen_dynamic_constant(*x, w)?;
                    } else {
                        write!(w, "::core::cmp::max(")?;
                        self.codegen_dynamic_constant(*x, w)?;
                        write!(w, ",")?;
                        opens += 1;
                    }
                }
                for _ in 0..opens {
                    write!(w, ")")?;
                }
            }
        }
        Ok(())
    }

    /*
     * Emit logic to index into an collection.
     */
    fn codegen_index_math(
        &self,
        mut collect_ty: TypeID,
        indices: &[Index],
    ) -> Result<String, Error> {
        let mut acc_offset = "0".to_string();
        for index in indices {
            match index {
                Index::Field(idx) => {
                    let Type::Product(ref fields) = self.module.types[collect_ty.idx()] else {
                        panic!()
                    };

                    // Get the offset of the field at index `idx` by calculating
                    // the product's size up to field `idx`, then offseting the
                    // base pointer by that amount.
                    for field in &fields[..*idx] {
                        let field_align = get_type_alignment(&self.module.types, *field);
                        let field = self.codegen_type_size(*field);
                        acc_offset = format!(
                            "((({} + {}) & !{}) + {})",
                            acc_offset,
                            field_align - 1,
                            field_align - 1,
                            field
                        );
                    }
                    let last_align = get_type_alignment(&self.module.types, fields[*idx]);
                    acc_offset = format!(
                        "(({} + {}) & !{})",
                        acc_offset,
                        last_align - 1,
                        last_align - 1
                    );
                    collect_ty = fields[*idx];
                }
                Index::Variant(idx) => {
                    // The tag of a summation is at the end of the summation, so
                    // the variant pointer is just the base pointer. Do nothing.
                    let Type::Summation(ref variants) = self.module.types[collect_ty.idx()] else {
                        panic!()
                    };
                    collect_ty = variants[*idx];
                }
                Index::Position(ref pos) => {
                    let Type::Array(elem, ref dims) = self.module.types[collect_ty.idx()] else {
                        panic!()
                    };

                    // The offset of the position into an array is:
                    //
                    //     ((0 * s1 + p1) * s2 + p2) * s3 + p3 ...
                    let elem_size = self.codegen_type_size(elem);
                    for (p, s) in zip(pos, dims) {
                        let p = self.get_value(*p);
                        acc_offset = format!("{} * ", acc_offset);
                        self.codegen_dynamic_constant(*s, &mut acc_offset)?;
                        acc_offset = format!("({} + {})", acc_offset, p);
                    }

                    // Convert offset in # elements -> # bytes.
                    acc_offset = format!("({} * {})", acc_offset, elem_size);
                    collect_ty = elem;
                }
            }
        }
        Ok(acc_offset)
    }

    /*
     * Lower the size of a type into a Rust expression.
     */
    fn codegen_type_size(&self, ty: TypeID) -> String {
        match self.module.types[ty.idx()] {
            Type::Control => panic!(),
            Type::Boolean | Type::Integer8 | Type::UnsignedInteger8 => "1".to_string(),
            Type::Integer16 | Type::UnsignedInteger16 => "2".to_string(),
            Type::Integer32 | Type::UnsignedInteger32 | Type::Float32 => "4".to_string(),
            Type::Integer64 | Type::UnsignedInteger64 | Type::Float64 => "8".to_string(),
            Type::Product(ref fields) => {
                let fields_align = fields
                    .into_iter()
                    .map(|id| get_type_alignment(&self.module.types, *id));
                let fields: Vec<String> = fields
                    .into_iter()
                    .map(|id| self.codegen_type_size(*id))
                    .collect();

                // Emit LLVM IR to round up to the alignment of the next field,
                // and then add the size of that field. At the end, round up to
                // the alignment of the whole struct.
                let mut acc_size = "0".to_string();
                for (field_align, field) in zip(fields_align, fields) {
                    acc_size = format!(
                        "(({} + {}) & !{})",
                        acc_size,
                        field_align - 1,
                        field_align - 1
                    );
                    acc_size = format!("({} + {})", acc_size, field);
                }
                let total_align = get_type_alignment(&self.module.types, ty);
                format!(
                    "(({} + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Summation(ref variants) => {
                let variants = variants.into_iter().map(|id| self.codegen_type_size(*id));

                // The size of a summation is the size of the largest field,
                // plus 1 byte and alignment for the discriminant.
                let mut acc_size = "0".to_string();
                for variant in variants {
                    acc_size = format!("::core::cmp::max({}, {})", acc_size, variant);
                }

                // No alignment is necessary before the 1 byte discriminant.
                let total_align = get_type_alignment(&self.module.types, ty);
                format!(
                    "(({} + 1 + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Array(elem, ref bounds) => {
                // The size of an array is the size of the element multipled by
                // the dynamic constant bounds.
                let mut acc_size = self.codegen_type_size(elem);
                for dc in bounds {
                    acc_size = format!("{} * ", acc_size);
                    self.codegen_dynamic_constant(*dc, &mut acc_size).unwrap();
                }
                format!("({})", acc_size)
            }
        }
    }

    /*
     * Generate a runner object for this function.
     */
    fn codegen_runner_object<W: Write>(&self, w: &mut W) -> Result<(), Error> {
        // Figure out the devices for the parameters and the return value if
        // they are collections and whether they should be immutable or mutable
        // references.
        let func = self.get_func();
        let mut param_devices = vec![None; func.param_types.len()];
        let mut return_device = None;
        for idx in 0..func.nodes.len() {
            match func.nodes[idx] {
                Node::Parameter { index } => {
                    let device = self.node_colors.get(&NodeID::new(idx));
                    assert!(param_devices[index].is_none() || param_devices[index] == device);
                    param_devices[index] = device;
                }
                Node::Return { control: _, data } => {
                    let device = self.node_colors.get(&data);
                    assert!(return_device.is_none() || return_device == device);
                    return_device = device;
                }
                _ => {}
            }
        }
        let mut param_muts = vec![false; func.param_types.len()];
        let mut return_mut = true;
        let objects = &self.collection_objects[&self.func_id];
        for idx in 0..func.param_types.len() {
            if let Some(object) = objects.param_to_object(idx)
                && objects.is_mutated(object)
            {
                param_muts[idx] = true;
            }
        }
        for object in objects.returned_objects() {
            if let Some(idx) = objects.origin(*object).try_parameter()
                && !param_muts[idx]
            {
                return_mut = false;
            }
        }

        // Emit the type definition. A runner object owns its backing memory.
        write!(
            w,
            "#[allow(non_camel_case_types)]\nstruct HerculesRunner_{} {{\n",
            func.name
        )?;
        for (device, _) in self.backing_allocations[&self.func_id].iter() {
            write!(w, "    backing_ptr_{}: *mut u8,\n", device.name(),)?;
            write!(w, "    backing_size_{}: usize,\n", device.name(),)?;
        }
        write!(w, "}}\n")?;
        write!(
            w,
            "impl HerculesRunner_{} {{\n    fn new() -> Self {{\n        Self {{\n",
            func.name
        )?;
        for (device, _) in self.backing_allocations[&self.func_id].iter() {
            write!(
                w,
                "            backing_ptr_{}: ::core::ptr::null_mut(),\n            backing_size_{}: 0,\n",
                device.name(),
                device.name()
            )?;
        }
        write!(w, "        }}\n    }}\n")?;
        write!(w, "    async fn run<'a>(&'a mut self")?;
        for idx in 0..func.num_dynamic_constants {
            write!(w, ", dc_p{}: u64", idx)?;
        }
        for idx in 0..func.param_types.len() {
            if self.module.types[func.param_types[idx].idx()].is_primitive() {
                write!(w, ", p{}: {}", idx, self.get_type(func.param_types[idx]))?;
            } else {
                let device = match param_devices[idx] {
                    Some(Device::LLVM) => "CPU",
                    Some(Device::CUDA) => "CUDA",
                    // For parameters that are unused, it doesn't really matter
                    // what device is required, so just pick CPU for now.
                    None => "CPU",
                    _ => panic!(),
                };
                let mutability = if param_muts[idx] { "Mut" } else { "" };
                write!(
                    w,
                    ", p{}: ::hercules_rt::Hercules{}Ref{}<'a>",
                    idx, device, mutability
                )?;
            }
        }
        if self.module.types[func.return_type.idx()].is_primitive() {
            write!(w, ") -> {} {{\n", self.get_type(func.return_type))?;
        } else {
            let device = match return_device {
                Some(Device::LLVM) => "CPU",
                Some(Device::CUDA) => "CUDA",
                _ => panic!(),
            };
            let mutability = if return_mut { "Mut" } else { "" };
            write!(
                w,
                ") -> ::hercules_rt::Hercules{}Ref{}<'a> {{\n",
                device, mutability
            )?;
        }
        write!(w, "    unsafe {{\n")?;
        for (device, (total, _)) in self.backing_allocations[&self.func_id].iter() {
            write!(w, "        let size = ")?;
            self.codegen_dynamic_constant(*total, w)?;
            write!(
                w,
                " as usize;\n        if self.backing_size_{} < size {{\n",
                device.name()
            )?;
            write!(w, "            ::hercules_rt::__{}_dealloc(self.backing_ptr_{}, self.backing_size_{});\n", device.name(), device.name(), device.name())?;
            write!(
                w,
                "            self.backing_size_{} = size;\n",
                device.name()
            )?;
            write!(w, "            self.backing_ptr_{} = ::hercules_rt::__{}_alloc(self.backing_size_{});\n", device.name(), device.name(), device.name())?;
            write!(w, "        }}\n")?;
        }
        for idx in 0..func.param_types.len() {
            if !self.module.types[func.param_types[idx].idx()].is_primitive() {
                write!(w, "        let p{} = p{}.__ptr();\n", idx, idx)?;
            }
        }
        write!(w, "        let ret = {}(", func.name)?;
        for (device, _) in self.backing_allocations[&self.func_id].iter() {
            write!(w, "self.backing_ptr_{}, ", device.name())?;
        }
        for idx in 0..func.num_dynamic_constants {
            write!(w, "dc_p{}, ", idx)?;
        }
        for idx in 0..func.param_types.len() {
            write!(w, "p{}, ", idx)?;
        }
        write!(w, ").await;\n")?;
        if self.module.types[func.return_type.idx()].is_primitive() {
            write!(w, "        ret\n")?;
        } else {
            let device = match return_device {
                Some(Device::LLVM) => "CPU",
                Some(Device::CUDA) => "CUDA",
                _ => panic!(),
            };
            let mutability = if return_mut { "Mut" } else { "" };
            write!(
                w,
                "        ::hercules_rt::Hercules{}Ref{}::__from_parts(ret, {} as usize)\n",
                device,
                mutability,
                self.codegen_type_size(func.return_type)
            )?;
        }
        write!(w, "    }}\n    }}\n")?;
        write!(
            w,
            "}}\nimpl Drop for HerculesRunner_{} {{\n    #[allow(unused_unsafe)]\n    fn drop(&mut self) {{\n        unsafe {{\n",
            func.name
        )?;
        for (device, _) in self.backing_allocations[&self.func_id].iter() {
            write!(
                w,
                "            ::hercules_rt::__{}_dealloc(self.backing_ptr_{}, self.backing_size_{});\n",
                device.name(),
                device.name(),
                device.name()
            )?;
        }
        write!(w, "        }}\n    }}\n}}\n")?;
        Ok(())
    }

    fn get_func(&self) -> &Function {
        &self.module.functions[self.func_id.idx()]
    }

    fn get_value(&self, id: NodeID) -> String {
        format!("node_{}", id.idx())
    }

    fn get_type(&self, id: TypeID) -> &'static str {
        convert_type(&self.module.types[id.idx()])
    }
}

fn convert_type(ty: &Type) -> &'static str {
    match ty {
        Type::Boolean => "bool",
        Type::Integer8 => "i8",
        Type::Integer16 => "i16",
        Type::Integer32 => "i32",
        Type::Integer64 => "i64",
        Type::UnsignedInteger8 => "u8",
        Type::UnsignedInteger16 => "u16",
        Type::UnsignedInteger32 => "u32",
        Type::UnsignedInteger64 => "u64",
        Type::Float32 => "f32",
        Type::Float64 => "f64",
        Type::Product(_) | Type::Summation(_) | Type::Array(_, _) => "*mut u8",
        _ => panic!(),
    }
}
